package binomialheap;

/**
 * A BinomialNode is a node in a binomial tree.
 *
 * A reference to a binomial node can be regarded as a reference to the binomial
 * tree consisting of its descendents.
 * If a tree has rank r then it contains 2<sup>r</sup> nodes.
 * A binomial tree is always a part of a circular list of trees which are in
 * strictly increasing rank order.
 *
 * The children of binomial node can be accessed through its child list trailer.
 * This is a reference to its highest rank child, acting as a trailer reference
 * to the start of the ordered list of children.
 * The priority of a child of a binomial node is no greater than the priority
 * of the node so the highest priority node in a binomial tree is always its
 * root.
 *
 * A list of binomial trees can be used to form a binomial heap.
 * In addition to its instance methods, the BinomialNode class provides some
 * static methods that are needed in the BinomialHeap class for implementing
 * its add() and remove() methods.
 */
public class BinomialNode
        implements Comparable<BinomialNode> {

  private int rank;
  private BinomialNode sibling;
  private BinomialNode childListTrailer;
  private Data data;

  /**
   * isValidListStartTrailer(listTrailer) returns true if listTrailer is a valid
   * list start trailer reference.
   * A valid list start trailer reference is either null or a reference to the
   * highest rank tree in a list with strictly increasing ranks.
   * @param startTrailer
   * @return
   */
  private static boolean isValidListStartTrailer(BinomialNode listTrailer) {
    if (listTrailer == null) {
      return true;
    }
    BinomialNode currentNode = listTrailer.sibling;
    int highestRank = currentNode.rank;
    currentNode = currentNode.sibling;
    while (currentNode != listTrailer) {
      int currentRank = currentNode.rank;
      if (currentRank <= highestRank) {
        return false;
      }
      highestRank = currentRank;
      currentNode = currentNode.sibling;
    }
    return true;
  }

  /**
   * tree.getTreeNodeCount() returns the number of binomial nodes in tree.
   * @return the number of nodes
   */
  private int getTreeNodeCount() {
    return 1 + getListNodeCount(childListTrailer);
  }

  /**
   * swapSiblings(tree1, tree2) swaps the siblings of tree1 and tree2.
   * This is a general purpose primitive operation that can be used to add a
   * tree into a tree list or remove a tree from a tree list.
   * If tree1 is its own sibling then swapSiblings(tree1, tree2) inserts tree1
   * into the tree list containing tree2.
   * It is inserted after tree2.
   * If tree1 is the sibling of tree2 in a list with more than one tree then
   * tree1 is removed from its tree list.
   * If tree1 and tree2 are the same tree then invoking this method has no
   * effect.
   * @param tree1 a tree
   * @param tree2 another tree
   */
  private static void swapSiblings(BinomialNode tree1, BinomialNode tree2) {
    // FIXME - add code.
  }

  /**
   * addTree(otherTree) adds otherTree at the end of the child list of this tree
   * incrementing the rank of this tree by one.
   * This method is intended as a helper method for mergeTrees().
   * Preconditions:
   * <ul>
   * <li>This tree and otherTree have the same rank.</li>
   * <li>The root of otherTree has lower priority than this tree.</li>
   * <li>This tree and otherTree both are in list containing just one tree.</li>
   * </ul>
   * Postconditions:
   * <ul>
   * <li>the rank has increased by 1</li>
   * </ul>
   * @param otherTree
   */
  private void addTree(BinomialNode otherTree) {
    assert otherTree.rank == rank :
            "Precondition violation: unmatched ranks";
    assert sibling == this :
            "Precondition violation: this tree is not a single tree";
    assert otherTree.sibling == otherTree :
            "Precondition violation: other tree is not a single tree";
    assert compareTo(otherTree) >= 0 :
            "Precondition violation: other tree has higher priority";
    // FIXME - add code.
    assert rank == otherTree.rank + 1 :
            "Postcondition violation: new rank is incorrect";
  }

  /**
   * mergeTrees(tree1, tree2) merges tree1 and tree2, returning the merged tree.
   * This method is intended as a helper method for mergeLists();
   * Preconditions:
   * <ul>
   * <li>tree1 and tree2 have the same rank</li>
   * <li>tree1 and tree2 are each the only trees in their tree list.</li>
   * </ul>
   * @param tree1 a binomial tree
   * @param tree2 another binomial tree with the same rank
   * @return the merged tree
   */
  private static BinomialNode mergeTrees(BinomialNode tree1, BinomialNode tree2) {
    // Note: preconditions are checked when addTree() is invoked.
    // FIXME - modify code.
    return tree2;
  }

  /**
   * BinomialNode.removeTree(trailer) removes the tree following trailer from
   * the tree list that contains trailer.
   * The returned value is a trailer reference suitable for manipulations
   * further along the list:
   * <ul>
   * <li>trailer if the list is nonempty after the remove operation</li>
   * <li>null if the list is empty after the remove.
   * </ul>
   * Preconditions:
   * <ul>
   * <li>trailer is not null</li>
   * </ul>
   * @param trailer : the trailer reference for the tree to be removed
   * @return a trailer reference for the list after the tree is removed
   */
  public static BinomialNode removeTree(BinomialNode trailer) {
    assert trailer != null :
            "trailer is null";
    // FIXME - modify code.
    return trailer;
  }

  /**
   * BinomialNode.mergeLists(trailer1, trailer2) merges the tree lists whose
   * start trailer references are trailer1 and trailer2 and returns the trailer
   * reference for the start of the merged list.
   * Preconditions:
   * <ul>
   * <li>trailer1 is not null</li>
   * </ul>
   * Preconditions:
   * <ul>
   * <li>trailer1 is a valid list start trailer reference</li>
   * <li>trailer2 is a valid list start trailer reference</li>
   * </ul>
   * Postconditions:
   * <ul>
   * <li>the return value is a valid list start trailer reference</li>
   * </ul>
   * @param trailer1 the trailer reference for the start of the first list
   * @param trailer2 the trailer reference for the start of the second list
   * @return the trailer reference for the start of the merged list
   */
  public static BinomialNode mergeLists(BinomialNode trailer1, BinomialNode trailer2) {
    assert isValidListStartTrailer(trailer1) :
            "Precondition violation: trailer1 is not a valid list start trailer";
    assert isValidListStartTrailer(trailer2) :
            "Precondition violation: trailer2 is not a valid list start trailer";
    // FIXME - first deal with empty lists (null trailers).
    BinomialNode currentList1Trailer = trailer1;
    // FIXME - add code.  Take special care.
    assert isValidListStartTrailer(currentList1Trailer) :
            "Postcondition violation: return value is not a valid list start trailer";
    return currentList1Trailer;
  }

  /**
   * BinomialTree.getHighestPriorityTrailer(startTrailer) returns a trailer
   * reference for the highest priority tree in the tree list whose start
   * trailer reference is startTrailer.
   * If hpt is the return value then hpt.getSibling is the highest priority tree
   * in the tree list whose trailer reference is listTrailer.
   * Preconditions:
   * <ul>
   * <li>startTrailer is a valid list start trailer reference</li>
   * </ul>
   * @param startTrailer a list start trailer reference for a tree list
   * @return a trailer reference for the highest priority tree in the list
   */
  public static BinomialNode getHighestPriorityTrailer(BinomialNode startTrailer) {
    assert isValidListStartTrailer(startTrailer) :
            "startTrailer is not a valid list start trailer";
    // FIXME - first deal with an empty list (null startTrailer).
    BinomialNode currentHighestPriorityTrailer = startTrailer;
    // FIXME - add code.  Take care.
    // Your code should be able to handle empty lists (null trailers).
    return currentHighestPriorityTrailer;
  }

  /**
   * new BinomialNode(data) returns a new rank 0 binomial node with data data.
   * A reference to this node can also be regarded as a reference to a binomial
   * tree or as a reference into a list of trees that are connected by their
   * sibling attributes.
   * @param data the data for the new node
   */
  public BinomialNode(Data data) {
    // FIXME - add code.
  }

  /**
   * BinomialTree.getListNodeCount(treeList) returns the number of nodes in the
   * trees in treeList.
   * @param treeList a tree list
   * @return the number of nodes
   */
  public static final int getListNodeCount(BinomialNode treeList) {
    if (treeList == null) {
      return 0;
    }
    BinomialNode currentTree = treeList.sibling;
    int count = currentTree.getTreeNodeCount();
    while (currentTree != treeList) {
      count += currentTree.getTreeNodeCount();
      currentTree = currentTree.sibling;
    }
    return count;
  }

  /**
   * node.compareTo(otherNode) returns a priority comparison between node and
   * otherNode.
   * This priority comparison is positive if the data in node has higher
   * priority than the data in otherNode, 0 if they have the same priority, or
   * negative if the data in node has lower priority than the data in otherNode.
   * @param otherNode
   * @return a priority comparison
   */
  public int compareTo(BinomialNode otherNode) {
    return data.compareTo(otherNode.data);
  }

  /**
   * node.getData() returns the data for node.
   * @return the data
   */
  public Data getData() {
    return data;
  }

  /**
   * tree.getSibling() returns the sibling of tree.
   * @return the sibling
   */
  public BinomialNode getSibling() {
    return sibling;
  }

  /**
   * node.getChildListTrailer() returns a trailer reference for the child list
   * of node.
   * @return the child list trailer
   */
  public BinomialNode getChildListTrailer() {
    return this.childListTrailer;
  }

  /**
   * node.toString() returns a string description for node.
   * This string specifies the rank and data for node.
   * @return a string description
   */
  @Override
  public String toString() {
    return "BinomialNode: rank = " + rank + ", data = " + data;
  }

}
